#include <string>
#include <iostream>
#include <set>
#include <vector>
using namespace std;

// #define DEBUG
// #define DEBUG2

// #define BENCHMARK 5000

// Define the pieces

// Helpers:
#define B0000 { 0, 0, 0, 0 }
#define B0001 { 0, 0, 0, 1 }
#define B0010 { 0, 0, 1, 0 }
#define B0011 { 0, 0, 1, 1 }
#define B0100 { 0, 1, 0, 0 }
#define B0101 { 0, 1, 0, 1 }
#define B0110 { 0, 1, 1, 0 }
#define B0111 { 0, 1, 1, 1 }
#define B1000 { 1, 0, 0, 0 }
#define B1001 { 1, 0, 0, 1 }
#define B1010 { 1, 0, 1, 0 }
#define B1011 { 1, 0, 1, 1 }
#define B1100 { 1, 1, 0, 0 }
#define B1101 { 1, 1, 0, 1 }
#define B1110 { 1, 1, 1, 0 }
#define B1111 { 1, 1, 1, 1 }

// There is one special piece that we handle seperatly:
// It's the piece that looks like a +.

const bool basic_pieces[][ 4 ][ 4 ][ 4 ] = {

    { { B1100, B0000, B0000, B0000 }, // 1488M, 3630:1
      { B0110, B0000, B0000, B0000 }, // 265s auf Athlon 1.4GHz, 1547 Lsg/s
      { B0010, B0000, B0000, B0000 }, // 326s auf P4-M 1.4GHz, 1258 Lsg/s
      { B0000, B0000, B0000, B0000 } },

    { { B1111, B0000, B0000, B0000 },
      { B1000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },
    
    { { B1000, B0000, B0000, B0000 },
      { B1100, B0000, B0000, B0000 },
      { B1000, B0000, B0000, B0000 },
      { B1000, B0000, B0000, B0000 } },

    { { B1100, B0000, B0000, B0000 },
      { B0110, B0000, B0000, B0000 },
      { B0100, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1110, B0000, B0000, B0000 },
      { B0100, B0000, B0000, B0000 },
      { B0100, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1111, B0000, B0000, B0000 },
      { B1010, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1110, B0000, B0000, B0000 },
      { B1100, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1100, B1000, B0000, B0000 },
      { B1100, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1100, B0000, B0000, B0000 },
      { B0100, B0000, B0000, B0000 },
      { B0110, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1100, B0000, B0000, B0000 },
      { B1000, B0000, B0000, B0000 },
      { B1100, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } },

    { { B1100, B0100, B0000, B0000 },
      { B1000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 },
      { B0000, B0000, B0000, B0000 } }
};

const int number_of_pieces = sizeof( basic_pieces ) / sizeof( basic_pieces[ 0 ] );

class piece;

// Handle the pieces in a 4x4x4 nested array and provide rotation.
// Additionally, aligns pieces to the lowest coordinates on request.
class cube
{
private:
    bool v[ 4 ][ 4 ][ 4 ];

public:
    explicit cube( const bool* p )
    {
        // Note that we swapped x and y on purpose.
        for( int z = 0; z < 4; ++z )
            for( int x = 0; x < 4; ++x )
                for( int y = 0; y < 4; ++y )
                    v[ x ][ y ][ z ] = *p++;
    }

    cube( const cube& src )
    {
        *this = src;
    }

    cube& operator=( const cube& src )
    {
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = src.v[ x ][ y ][ z ];
        return *this;
    }

    void rotate_x()
    {
        cube cpy( *this );
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = cpy.v[ x ][ z ][ 3-y ];
    }

    void rotate_y()
    {
        cube cpy( *this );
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = cpy.v[ z ][ y ][ 3-x ];
    }

    void rotate_z()
    {
        cube cpy( *this );
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = cpy.v[ y ][ 3-x ][ z ];
    }

private:
    bool empty_x() const
    {
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                if( v[ 0 ][ y ][ z ] )
                    return false;
        return true;
    }

    bool empty_y() const
    {
        for( int z = 0; z < 4; ++z )
            for( int x = 0; x < 4; ++x )
                if( v[ x ][ 0 ][ z ] )
                    return false;
        return true;
    }

    bool empty_z() const
    {
        for( int y = 0; y < 4; ++y )
            for( int x = 0; x < 4; ++x )
                if( v[ x ][ y ][ 0 ] )
                    return false;
        return true;
    }

    void shift_x()
    {
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 3; ++x )
                    v[ x ][ y ][ z ] = v[ x + 1 ][ y ][ z ];
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                v[ 3 ][ y ][ z ] = false;
    }

    void shift_y()
    {
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 3; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = v[ x ][ y + 1 ][ z ];
        for( int z = 0; z < 4; ++z )
            for( int x = 0; x < 4; ++x )
                v[ x ][ 3 ][ z ] = false;
    }

    void shift_z()
    {
        for( int z = 0; z < 3; ++z )
            for( int y = 0; y < 4; ++y )
                for( int x = 0; x < 4; ++x )
                    v[ x ][ y ][ z ] = v[ x ][ y ][ z + 1 ];
        for( int y = 0; y < 4; ++y )
            for( int x = 0; x < 4; ++x )
                v[ x ][ y ][ 3 ] = false;
    }

public:
    void align()
    {
        while( empty_x() )
            shift_x();
        while( empty_y() )
            shift_y();
        while( empty_z() )
            shift_z();
    }

    bool valid() const
    {
        for( int z = 0; z < 4; ++z )
            for( int y = 0; y < 4; ++y )
                if( v[ 3 ][ y ][ z ] )
                    return false;
        return true;
    }

    friend class piece;
};

struct piece
{
    unsigned int a, b;
    explicit piece( const unsigned int va = 0, const unsigned int vb = 0 ) : a( va ), b( vb ) {}
    explicit piece( const cube& c ) : a( 0 ), b( 0 )
    {
        for( int z = 0; z < 4; ++z ) {
            for( int y = 0; y < 4; ++y ) {
                for( int x = 0; x < 3; ++x ) {
                    if( c.v[ x ][ y ][ z ] ) {
                        const int bit = x + y * 3 + z * 12;
                        if( bit < 28 )
                            a |= 1 << bit;
                        else
                            b |= 1 << ( bit - 28 );
                    }
                }
            }
        }
    }

    piece rotated( const bool rx = false, const bool ry = false, const bool rz = false ) const
    {
        piece result;
        for( int x = 0; x < 3; ++x ) {
            for( int y = 0; y < 4; ++y ) {
                for( int z = 0; z < 5; ++z ) {
                    const int sourceBit = x + y * 3 + z * 12;
                    if( ( sourceBit < 28 ) ? ( a & ( 1 << sourceBit ) ) : ( b & ( 1 << sourceBit - 28 ) ) ) {
                        const int destBit = ( rx ? x : ( 2 - x ) ) + ( ry ? y : ( 3 - y ) ) * 3 + ( rz ? z : ( 4 - z ) ) * 12;
                        if( destBit < 28 )
                            result.a |= 1 << destBit;
                        else
                            result.b |= 1 << ( destBit - 28 );
                    }
                }
            }
        }
        return result;
    }
};

bool operator==( const piece& lhs, const piece& rhs )
{
    return lhs.a == rhs.a && lhs.b == rhs.b;
}

bool operator<( const piece& lhs, const piece& rhs )
{
    return lhs.a < rhs.a || ( lhs.a == rhs.a && lhs.b < rhs.b );
}

inline int first_hole( const unsigned int b, int i )
{
    while( b & ( 1 << i ) )
        ++i;
    
    return i;
}
 
inline int first_hole( const unsigned int a, const unsigned int b, const int i )
{
    const int t = ( i < 28 ) ? first_hole( a, i ) : i;
    return ( t < 28 ) ? t : first_hole( b, t - 28 ) + 28;
}

int first_bit( const piece& p )
{
    return first_hole( ~p.a & 0xFFFFFFF, ~p.b, 0 );
}

// Note that using a set eliminates duplicates automatically :)
set< piece > rotated( const cube& c )
{
    set< piece > result;
    for( int rx = 0; rx < 4; ++rx ) {
        for( int ry = 0; ry < 4; ++ry ) {
            for( int rz = 0; rz < 4; ++rz ) {
                cube tmp( c );
                for( int i = 0; i < rx; ++i )
                    tmp.rotate_x();
                for( int i = 0; i < ry; ++i )
                    tmp.rotate_y();
                for( int i = 0; i < rz; ++i )
                    tmp.rotate_z();
                tmp.align();
                if( tmp.valid() )
                    result.insert( piece( tmp ) );
            }
        }
    }
    return result;
}

bool empty_x( const piece& p )
{
    return !( p.a & 0x4924924 ) && !( p.b & 0x92492492 );
}

bool empty_y( const piece& p )
{
    return !( p.a & 0x0E00E00 ) && !( p.b & 0xE00E00E0 );
}

bool empty_z( const piece& p )
{
    return !( p.b & 0xFFF00000 );
}

template< int I > piece shift( const piece& p )
{
    return piece( ( p.a << I ) & 0xFFFFFFF, ( p.b << I ) | ( p.a >> ( 28 - I ) ) );
}

// Now, we add all pieces at all possible positions
set< piece > moved( set< piece > result )
{
    while( true ) {
        const size_t s = result.size();
        set< piece > tmp;
        for( set< piece >::const_iterator it = result.begin(); it != result.end(); ++it ) {
            if( empty_x( *it ) )
                tmp.insert( shift< 1 >( *it ) );
            if( empty_y( *it ) )
                tmp.insert( shift< 3 >( *it ) );
            if( empty_z( *it ) )
                tmp.insert( shift< 12 >( *it ) );
        }
        result.insert( tmp.begin(), tmp.end() );
        if( result.size() == s )
            return result;
    }
}

// All functions above are only used to initialize a single map:
// The map of all possible pieces in all possible positions
// and with some additional information about the first bit.
// This map is created here and solve() uses nothing but this
// map and the piece-structure itself. The above functions are
// not really optimized for speed, as the real work is done by
// the function [q]step() anyway.

vector< piece > basic_global_map[ 60 ][ number_of_pieces ];

vector< piece > global_map[ 60 ][ number_of_pieces ];
vector< unsigned int > qglobal_map[ 32 ][ number_of_pieces ];

void fill_basic_map()
{
    for( unsigned int i = 0; i < number_of_pieces; ++i ) {
        const cube c( &basic_pieces[ i ][ 0 ][ 0 ][ 0 ] );
        const set< piece >& p = moved( rotated( c ) );
        for( set< piece >::const_iterator it = p.begin(); it != p.end(); ++it )
            basic_global_map[ first_bit( *it ) ][ i ].push_back( *it );
    }
}

void fill_map( const piece& limit )
{
    for( unsigned int b = 0; b < 60; ++b )
        for( unsigned int i = 0; i < number_of_pieces; ++i )
            global_map[ b ][ i ].clear();
    for( unsigned int b = 0; b < 32; ++b )
        for( unsigned int i = 0; i < number_of_pieces; ++i )
            qglobal_map[ b ][ i ].clear();

    for( unsigned int b = 0; b < 60; ++b )
        for( unsigned int i = 0; i < number_of_pieces; ++i )
            for( vector< piece >::const_iterator it = basic_global_map[ b ][ i ].begin(); it != basic_global_map[ b ][ i ].end(); ++it )
                if( !( it->a & limit.a ) && !( it->b & limit.b ) )
                    global_map[ first_bit( *it ) ][ i ].push_back( *it );
    
    for( unsigned int b = 0; b < 60; ++b )
        for( unsigned int i = 0; i < number_of_pieces; ++i )
            if( global_map[ b ][ i ].empty() )
                global_map[ b ][ i ].push_back( piece( ~0, ~0 ) );
  
    for( unsigned int b = 0; b < 32; ++b )
        for( unsigned int i = 0; i < number_of_pieces; ++i )
            for( vector< piece >::const_iterator it = global_map[ b + 28 ][ i ].begin(); it != global_map[ b + 28 ][ i ].end(); ++it )
                qglobal_map[ b ][ i ].push_back( it->b );
}

vector< piece > replace_first()
{
    // The first piece is special. In order to detect symmetric
    // solutions, we just limit the first piece to the right
    // directions / positions.
    vector< piece > result;

    cube c( &basic_pieces[ 0 ][ 0 ][ 0 ][ 0 ] );
    const set< piece >& p = moved( rotated( c ) );
    set< piece > s;
    for( set< piece >::const_iterator it = p.begin(); it != p.end(); ++it ) {
        if( s.find( *it ) != s.end() )
            continue;

        const piece rot = it->rotated( false, true, false );

        const bool b = first_bit( *it ) < first_bit( rot );
        result.push_back( b ? *it : rot );

        s.insert( *it );
        s.insert( rot );
    }

    return result;
}

void show_map()
{
    int t = 0;
    for( int i = 0; i < 60; ++i ) {
        for( int j = 0; j < number_of_pieces; ++j ) {
            t += global_map[ i ][ j ].size();
            cout << global_map[ i ][ j ].size() << '\t';
        }
        cout << '\n';
        if( i == 27 )
            cout << '\n';
    }
    cout << "\nTotal: " << t << endl;
}

// Now for the main algorithm to find / count the solutions:

unsigned int counter = 0;

#ifdef DEBUG2
unsigned long long step_counter = 0;
unsigned long long qstep_counter = 0;
#endif

#ifdef DEBUG
unsigned long long placed = 0;

void stats()
{
    cout << '\r'
#ifdef DEBUG2
         << step_counter << "\r\t\t" << qstep_counter << "\r\t\t\t\t"
#endif
         << ( placed / 1000000 ) << "M   " << counter << "   ( "
         << int( double( placed ) / counter ) << " : 1 ) " << flush;
}
#endif

void qstep( const unsigned int b, const unsigned int allocated, int first_hole_test )
{
#ifdef DEBUG2
    ++qstep_counter;
#endif

    first_hole_test = first_hole( b, ++first_hole_test );

    const vector< unsigned int >* g = &qglobal_map[ first_hole_test ][ 0 ];
    unsigned int i = 1;
    do {
        if( allocated & i ) {
            vector< unsigned int >::const_iterator p = g->begin();
            do {
                if( !( b & *p ) ) {
                    if( ~( b | *p ) ) {
#ifdef DEBUG
                        if( !( ++placed & 0x3FFFFF ) )
                            stats();
#endif
                        qstep( b | *p, allocated ^ i, first_hole_test );
                    }
                    else {
                        ++counter;
#ifdef BENCHMARK
                        if( counter > BENCHMARK )
                            exit( 0 );
#endif
                    }
                }
            } while( ++p != g->end() );
        }
        ++g;
    } while( ( i <<= 1 ) != 1 << number_of_pieces );
}

void step( const unsigned int a, const unsigned int b, const unsigned int allocated, int first_hole_test = -1 )
{
#ifdef DEBUG2
    ++step_counter;
#endif

    first_hole_test = first_hole( a, b, ++first_hole_test );
    const vector< piece >* g = &global_map[ first_hole_test ][ 0 ];
    unsigned int i = 1;
    do {
        if( allocated & i ) {
            vector< piece >::const_iterator p = g->begin();
            do {
                if( !( b & p->b ) && !( a & p->a ) ) {
                    if( ( a | p->a ) == 0xFFFFFFF ) {
                        if( !( allocated ^ i ) ) {
                            ++counter;
#ifdef BENCHMARK
                            if( counter > BENCHMARK )
                                exit( 0 );
#endif
                        }
                        else {
#ifdef DEBUG
                            if( !( ++placed & 0x3FFFFF ) )
                                stats();
#endif
                            qstep( b | p->b, allocated ^ i, std::max( -1, first_hole_test - 28 ) );
                        }
                    }
                    else {
#ifdef DEBUG
                        if( !( ++placed & 0x3FFFFF ) )
                            stats();
#endif
                        step( a | p->a, b | p->b, allocated ^ i, first_hole_test );
                    }
                }
            } while( ++p != g->end() );
        }
        ++g;
    } while( ( i <<= 1 ) != 1 << number_of_pieces );
}

int main()
{
    fill_basic_map();

    // The '+'-piece in all possible positions, reduced by the symmetric positions
    piece p( 0xba, 0 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0xba000, 0 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0x2007002, 0 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0x38010, 1 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0x8049008, 0 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0x92010, 1 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    p = piece( 0x124020, 2 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );
    
    p = piece( 0x9008000, 0x804 );
    fill_map( p );
    step( p.a, p.b, 0x7FF );

    // Some positions are special as they are symmetric at the x-axis.
    // Use the first piece of the others to limit the search.
    const vector< piece >& v = replace_first();
    
    for( vector< piece >::const_iterator it = v.begin(); it != v.end(); ++it ) {
        piece p( 0xa000000, 0xb );
        if( !( p.a & it->a ) && !( p.b & it->b ) ) {
            fill_map( piece( p.a | it->a, p.b | it->b ) );
            step( p.a | it->a, p.b | it->b, 0x7FE );
        }
        
        p = piece( 0x7002000, 0x200 );
        if( !( p.a & it->a ) && !( p.b & it->b ) ) {
            fill_map( piece( p.a | it->a, p.b | it->b ) );
            step( p.a | it->a, p.b | it->b, 0x7FE );
        }
        
        p = piece( 0x8010000, 0x1003 );
        if( !( p.a & it->a ) && !( p.b & it->b ) ) {
            fill_map( piece( p.a | it->a, p.b | it->b ) );
            step( p.a | it->a, p.b | it->b, 0x7FE );
        }
            
        p = piece( 0x2010000, 0x1009 );
        if( !( p.a & it->a ) && !( p.b & it->b ) ) {
            fill_map( piece( p.a | it->a, p.b | it->b ) );
            step( p.a | it->a, p.b | it->b, 0x7FE );
        }
    }

#ifdef DEBUG
    stats();
#endif

    cout << "\n\nSolutions: " << counter << endl;
}
